from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import List

import cv2
from facenet_pytorch.models.mtcnn import MTCNN
from PIL import Image
from torch.utils.data.dataloader import DataLoader, Dataset


class VideoFaceDetector(ABC):

    def __init__(self, **kwargs) -> None:
        super().__init__()

    @property
    @abstractmethod
    def _batch_size(self) -> int:
        pass

    @abstractmethod
    def _detect_faces(self, frames) -> List:
        pass


class FacenetDetector(VideoFaceDetector):

    def __init__(self, device="cuda:0") -> None:
        super().__init__()

        self.detector = MTCNN(
            device=device,
            thresholds=[0.85, 0.95, 0.95],
            margin=0,
        )

    def _detect_faces(self, frames) -> List:
        """
        passes all the frames of a video and then returns the bboxes
        of each frame into batch_boxes.
        """
        batch_boxes, *_ = self.detector.detect(frames, landmarks=False)
        if batch_boxes is None:
            return []
        return [b.tolist() if b is not None else None for b in batch_boxes]

    @property
    def _batch_size(self):
        return 32



from collections import OrderedDict

import cv2
from PIL import Image
from torch.utils.data.dataloader import Dataset


class VideoDataset(Dataset):
    """
    Dataset class for fetching specific information about a video.
    The class requires a list of videos that it has to process.
    """

    def __init__(self, videos) -> None:
        super().__init__()
        self.videos = videos

    def __getitem__(self, index: int):
        """
        This function, picks a video and returns 4 values.
            str: Contains the name of the video
            list: List containing only the frame indices that are readable
            number: fps of the video
            list: a list containing all the frames as image arrays
        """
        video = self.videos[index]
        capture = cv2.VideoCapture(video)
        frames_num = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
        fps = int(capture.get(5))
        frames = OrderedDict()
        for i in range(frames_num):
            capture.grab()
            success, frame = capture.retrieve()
            if not success:
                continue
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            frame = Image.fromarray(frame)
            frame = frame.resize(size=[s // 2 for s in frame.size])
            frames[i] = frame
        return video, list(frames.keys()), fps, list(frames.values())

    def __len__(self) -> int:
        return len(self.videos)



from typing import Dict, List

import networkx as nx
import numpy as np
import torch
from facenet_pytorch import InceptionResnetV1, fixed_image_standardization
from torch.utils.data.dataloader import DataLoader
from torchvision.transforms import Resize

from identity_clustering.dataset import VideoDataset
from identity_clustering.facedetector import FacenetDetector
from identity_clustering.utils import extract_crops


class FaceCluster:

    def __init__(
        self, crops=None, similarity_threshold: int = 0.85, device: str = "cpu"
    ):
        self.similarity_threshold = similarity_threshold
        self.device = torch.device(device)
        self.crops = crops

    def _set_crops(self, crops):
        self.crops = crops

    def _set_threshold(self, threshold):
        self.similarity_threshold = threshold

    def _preprocess_images(self, img, shape=[128, 128]):
        img = Resize(shape)(img)
        return img

    def _generate_connected_components(self, similarities):

        graph = nx.Graph()
        for i in range(len(similarities)):
            for j in range(len(similarities)):
                # take every face dot product value
                # with every other value except itself and compare with threshold, when
                # if the value is greater than the threshold then
                # add an edge between them to signify that they are the same face.
                if i != j and similarities[i, j] > self.similarity_threshold:
                    graph.add_edge(i, j)
        # get all the clustered components and add them to the resultant component list
        components_list = []
        for component in nx.connected_components(graph):
            components_list.append(list(component))
        # for memory optimization clear the graph that was created
        graph.clear()
        graph = None

        return components_list

    def cluster_faces(self, crops, threshold=None):

        if threshold:
            self._set_threshold(threshold)

        if crops:
            self._set_crops(crops)

        # Convert crops to PIL images
        crops_images = [row[3] for row in self.crops]

        # Extract the embeddings
        embeddings_extractor = (
            InceptionResnetV1(pretrained="vggface2").eval().to(self.device)
        )
        faces = [self._preprocess_images(face) for face in crops_images]
        faces = np.stack([np.uint8(face) for face in faces])
        faces = torch.as_tensor(faces)
        faces = faces.permute(0, 3, 1, 2).float()
        faces = fixed_image_standardization(faces)
        face_recognition_input = faces
        embeddings = []
        embeddings = embeddings_extractor(face_recognition_input.to(self.device))
        similarities = (
            torch.tensordot(embeddings, embeddings.T, dims=1).detach().cpu().numpy()
        )
        # dot product attention without scaling
        # similarities = np.dot(np.array(embeddings), np.array(embeddings).T)

        # use the helper function to generate clusters
        components = self._generate_connected_components(similarities)
        components = [sorted(component) for component in components]

        # assigning each cluster to a unique identity.
        clustered_faces = {}
        for identity_index, component in enumerate(components):
            for index, face_index in enumerate(component):
                component[index] = self.crops[face_index]

            clustered_faces[identity_index] = component

        return clustered_faces


def cluster(
    clust: FaceCluster,
    video_path: str,
    faces: List[tuple],
    pad_constant: int | tuple | None = 3,
) -> Dict[int, list]:
    crops = extract_crops(video_path, faces, pad_constant)
    clustered_faces = clust.cluster_faces(crops)
    return clustered_faces


def detect_faces(video_path, device):
    """
    video_path: str - Path to the video
    device: str - indicates whether to leverage CPU or GPU for processing
    """

    """
        We'll be using the facenet detector that is required to detect the faces
        present in each frame. This function is only responsible to return
        a dictionary that contains the bounding boxes of each frame.
        returns: 
            dict: dict template:
                {
                    frame_no: [[
                        [number, number, number, number],
                        [number, number, number, number],
                        ...
                        ...
                        [number, number, number, number]
                    ]]
                }
            int: fps of the video
    """
    detector = FacenetDetector(device=device)

    # Read the video and its information
    dataset = VideoDataset([video_path])
    loader = DataLoader(
        dataset, shuffle=False, num_workers=0, batch_size=1, collate_fn=lambda x: x
    )

    # Detect the faces
    for item in loader:
        bboxes = {}
        video, indices, fps, frames = item[0]
        """
            Update bboxes dict with the bounding boxes present in each frame with the 
            frame number as the index and 
            a two dimensional list containing the bounding boxes as the value. 
        """
        bboxes.update({i: b for i, b in zip(indices, detector._detect_faces(frames))})
        found_faces = False
        for key in list(bboxes.keys()):
            if isinstance(bboxes[key], list):
                found_faces = True
                break

        if not found_faces:
            return None, indices[-1]
    return bboxes, fps



import cv2
from PIL import Image


def _get_frames(video_path):
    """
    This function gets the video path, reads the video, stores
    the frames in a list and then returns
    """

    # List to store the video frames
    frames = []

    # Read and store video Frames
    capture = cv2.VideoCapture(video_path)
    frames_num = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))

    for _ in range(frames_num):
        capture.grab()
        success, frame = capture.retrieve()
        if not success:
            continue
        frames.append(frame)
    return frames


def _get_crop(frame, bbox, pad_constant: int | tuple):
    """
    This function takes a frame and a bounding box and outputs the region of the image
    given by the bounding box with padding applied to all four sides.

    Args:
        - frame (np.ndarray): The image frame containing the faces to be cropped.
        - bbox (list): The bounding box (xmin, ymin, xmax, ymax) for cropping.
        - pad_constant (int | tuple): The constant to control the padding.
          If an integer is provided, the same padding is applied in all directions.
          If a tuple (pad_w, pad_h) is provided, different padding is
          applied to width and height.

    Returns:
        - crop (np.ndarray): The cropped face region from the frame with padding.
    """

    # Extract bounding box coordinates
    xmin, ymin, xmax, ymax = [int(b * 2) for b in bbox]

    # Handle padding logic
    if isinstance(pad_constant, int):
        p_w = p_h = pad_constant
    elif isinstance(pad_constant, tuple) and len(pad_constant) == 2:
        p_w, p_h = pad_constant
    else:
        raise ValueError(
            "pad_constant should be either an int or a tuple of two values."
        )

    # Define padded crop area
    crop_ymin = max(ymin - p_h, 0)  # Ensure it doesn't go below 0
    crop_xmin = max(xmin - p_w, 0)  # Ensure it doesn't go left of 0
    crop_ymax = min(
        ymax + p_h, frame.shape[0]
    )  # Ensure it doesn't go beyond image height
    crop_xmax = min(
        xmax + p_w, frame.shape[1]
    )  # Ensure it doesn't go beyond image width

    # Extract the cropped region from the frame
    crop = frame[crop_ymin:crop_ymax, crop_xmin:crop_xmax]

    return crop


def extract_crops(video_path, bboxes_dict, pad_constant: int | tuple = 50):

    frames = _get_frames(video_path)
    crops = []
    keys = [int(x) for x in list(bboxes_dict.keys())]
    for i in range(0, len(frames)):
        frame = frames[i]
        if i not in keys:
            continue
        bboxes = bboxes_dict[i]
        if not bboxes:
            continue
        for bbox in bboxes:
            crop = _get_crop(frame, bbox, pad_constant)
            fram = _get_crop(frame, bbox, 0)

            # Add the extracted face to the list
            crops.append((i, Image.fromarray(crop), bbox, Image.fromarray(fram)))

    return crops
